#!/usr/bin/env python3
"""Use SIFT to register an image to previous ones.

This script initially worked on a pair of subsequent images.
Then I changed approach and rewrote its core to be a function to be
used by other scripts.
So I do not know whether the original part (the one inside the
if __name__ == '__main__') still works.

Many thanks to https://datahacker.rs/feature-matching-methods-comparison-in-opencv/
for the refresher on SIFT.

Written in November 2021 by
Pier Angelo Vendrame <vogliadifarniente@gmail.com>.

To the extent possible under law, the author has dedicated all
copyright and related and neighboring rights to this software to the
public domain worldwide. This software is distributed without any
warranty.
"""

from glob import glob
import time

import cv2
import numpy as np

show = True
sift_points = 20000
ransac_error = 2.0
print_msgs = False


def find_homography(im, targets):
    sift = cv2.SIFT_create(sift_points)
    t = time.time()
    kp, des = sift.detectAndCompute(im, None)

    kps = []
    dess = []
    for t in targets:
        kpt, dest = sift.detectAndCompute(t, None)
        kps.append(kpt)
        dess.append(dest)
    if print_msgs:
        print('SIFT time', time.time() - t)

    # https://datahacker.rs/feature-matching-methods-comparison-in-opencv/
    bf = cv2.BFMatcher()
    start_t = time.time()
    good_matches = []
    tot_matches = 0
    for dest in dess:
        good_matches.append([])
        matches = bf.knnMatch(dest, des, k=2)
        if print_msgs:
            print('Match time', time.time() - start_t)
        for m1, m2 in matches:
            if m1.distance < 0.6 * m2.distance:
                good_matches[-1].append([m1])
        if print_msgs:
            print('Good matches', len(good_matches))
        tot_matches += len(good_matches[-1])

    src_points = np.empty((tot_matches, 2))
    dst_points = np.empty(src_points.shape)
    offset = 0
    for t, gm in enumerate(good_matches):
        for i, m in enumerate(gm):
            # https://stackoverflow.com/questions/30716610/how-to-get-pixel-coordinates-from-feature-matching-in-opencv-python
            src_points[offset + i] = kp[m[0].trainIdx].pt
            dst_points[offset + i] = kps[t][m[0].queryIdx].pt
        offset += len(gm)
    H, mask = cv2.findHomography(
        src_points, dst_points, cv2.RANSAC, ransac_error)

    if show:
        for t, imt in enumerate(targets):
            im3 = cv2.drawMatchesKnn(
                imt, kps[t], im, kp, good_matches[t], None, flags=2)
            cv2.namedWindow('Matches', cv2.WINDOW_NORMAL)
            cv2.imshow('Matches', im3)

            im2 = cv2.warpPerspective(im, H, (imt.shape[1], imt.shape[0]))
            cv2.namedWindow('Im1', cv2.WINDOW_NORMAL)
            cv2.imshow('Im1', imt)
            cv2.namedWindow('Im2', cv2.WINDOW_NORMAL)
            cv2.imshow('Im2', im2)

            im4 = cv2.addWeighted(imt, 0.5, im2, 0.5, 0)
            cv2.namedWindow('Overlapped', cv2.WINDOW_NORMAL)
            cv2.imshow('Overlapped', im4)

            cv2.waitKey(0)

    return H


if __name__ == '__main__':
    print_msgs = True
    frames = sorted(glob('zoom/*.jpg'))
    Hs = np.empty((len(frames) - 1, 3, 3))
    for i in range(len(frames) - 1):
        print('Starting', i)
        im1 = cv2.imread(frames[i])
        im2 = cv2.imread(frames[i + 1])
        Hs[i] = find_homography(im2, [im1])
        np.save(f'H_{sift_points}.npy', Hs)
